﻿namespace System.Numerics.Geometry
{
	using System.Collections.Generic;
	using System.Diagnostics;
	using System.Globalization;
	using System.Runtime.InteropServices;
	using System.Xml;
	using System.Xml.Serialization;

	/// <summary>
	/// Representation of a plane.
	/// </summary>
	[Serializable]
	[DebuggerDisplay("A = {A} B = {B} C = {C} D = {D}")]
	[StructLayout(LayoutKind.Sequential, Pack = 1)]
	public struct Plane : IEnumerable<Single>, IEquatable<Plane>, IFormattable, IXmlSerializable
	{
		#region Constructors
		/// <summary>
		/// Initializes a new instance of the struct.
		/// </summary>
		/// <param name="abcd">The value for all the components.</param>
		public Plane(Single abcd)
		{
			this.A = abcd;
			this.B = abcd;
			this.C = abcd;
			this.D = abcd;
		}

		/// <summary>
		/// Initializes a new instance of the struct.
		/// </summary>
		/// <param name="normal">The plane normal.</param>
		/// <param name="d">The D component of the plane equation.</param>
		public Plane(Vector3 normal, Single d)
		{
			this.A = normal.X;
			this.B = normal.Y;
			this.C = normal.Z;
			this.D = d;
		}

		/// <summary>
		/// Initializes a new instance of the struct.
		/// </summary>
		/// <param name="normal">The plane normal.</param>
		/// <param name="pointOnPlane">A point on the plane.</param>
		public Plane(Vector3 normal, Vector3 pointOnPlane)
		{
			this.A = normal.X;
			this.B = normal.Y;
			this.C = normal.Z;
			this.D = -Vector3.Dot(normal, pointOnPlane);
		}

		/// <summary>
		/// Initializes a new instance of the struct.
		/// </summary>
		/// <param name="a">The A component of the plane equation.</param>
		/// <param name="b">The B component of the plane equation.</param>
		/// <param name="c">The C component of the plane equation.</param>
		/// <param name="d">The D component of the plane equation.</param>
		public Plane(Single a, Single b, Single c, Single d)
		{
			this.A = a;
			this.B = b;
			this.C = c;
			this.D = d;
		}
		#endregion

		#region Methods
		/// <summary>
		/// Determines whether the current object is equal to the specified object.
		/// </summary>
		/// <param name="value">An object.</param>
		/// <returns>True if the current object is equal to the <paramref name="value"/>, otherwise false.</returns>
		public override bool Equals(object value)
		{
			if (value is Plane)
			{
				return Equals((Plane)value);
			}

			return false;
		}

		/// <summary>
		/// Determines whether the current plane is equal to the specified plane.
		/// </summary>
		/// <param name="value">A plane.</param>
		/// <returns>True if the current plane is equal to the <paramref name="value"/>, otherwise false.</returns>
		public bool Equals(Plane value)
		{
			return
				this.A == value.A &&
				this.B == value.B &&
				this.C == value.C &&
				this.D == value.D;
		}

		/// <summary>
		/// Returns the plane created from the specified points.
		/// </summary>
		/// <param name="point1">A point.</param>
		/// <param name="point2">A point.</param>
		/// <param name="point3">A point.</param>
		/// <returns>The plane created from the <paramref name="point1"/>, <paramref name="point2"/> and the <paramref name="point3"/>.</returns>
		public static Plane FromPoints(Vector3 point1, Vector3 point2, Vector3 point3)
		{
			return new Plane(Vector3.Normalize(Vector3.Cross(point2 - point1, point3 - point1)), point1);
		}

		/// <summary>
		/// Returns a hash code for the current plane.
		/// </summary>
		/// <returns>A hash code.</returns>
		/// <remarks>
		/// The hash code is not unique.
		/// If two planes are equal, their hash codes are guaranteed to be equal.
		/// If the planes are not equal, their hash codes are not guaranteed to be different.
		/// </remarks>
		public override int GetHashCode()
		{
			return
				this.A.GetHashCode() ^
				this.B.GetHashCode() ^
				this.C.GetHashCode() ^
				this.D.GetHashCode();
		}

		/// <summary>
		/// Returns an enumerator that iterates through the elements.
		/// </summary>
		/// <returns>
		/// An enumerator object that can be used to iterate through the elements.
		/// </returns>
		public IEnumerator<Single> GetEnumerator()
		{
			yield return this.A;
			yield return this.B;
			yield return this.C;
			yield return this.D;
		}

		/// <summary>
		/// Returns an enumerator that iterates through the elements.
		/// </summary>
		/// <returns>
		/// An enumerator object that can be used to iterate through the elements.
		/// </returns>
		Collections.IEnumerator Collections.IEnumerable.GetEnumerator()
		{
			return GetEnumerator();
		}

		/// <summary>
		/// Determines whether any of the components of the specified plane evaluates to an infinity.
		/// </summary>
		/// <param name="value">A plane.</param>
		/// <returns>True if any of the components of <paramref name="value"/> evaluates to an infinity, otherwise false.</returns>
		public static bool IsInfinity(Plane value)
		{
			return
				Single.IsInfinity(value.A) ||
				Single.IsInfinity(value.B) ||
				Single.IsInfinity(value.C) ||
				Single.IsInfinity(value.D);
		}

		/// <summary>
		/// Determines whether any of the components of the specified plane evaluates to a value that is not a number.
		/// </summary>
		/// <param name="value">A plane.</param>
		/// <returns>True if any of the components of <paramref name="value"/> evaluates to a value that is not a number, otherwise false.</returns>
		public static bool IsNaN(Plane value)
		{
			return
				Single.IsNaN(value.A) ||
				Single.IsNaN(value.B) ||
				Single.IsNaN(value.C) ||
				Single.IsNaN(value.D);
		}

		/// <summary>
		/// Determines whether any of the components of the specified plane evaluates to negative infinity.
		/// </summary>
		/// <param name="value">A plane.</param>
		/// <returns>True if any of the components of <paramref name="value"/> evaluates to negative infinity, otherwise false.</returns>
		public static bool IsNegativeInfinity(Plane value)
		{
			return
				Single.IsNegativeInfinity(value.A) ||
				Single.IsNegativeInfinity(value.B) ||
				Single.IsNegativeInfinity(value.C) ||
				Single.IsNegativeInfinity(value.D);
		}

		/// <summary>
		/// Determines whether any of the components of the specified plane evaluates to positive infinity.
		/// </summary>
		/// <param name="value">A plane.</param>
		/// <returns>True if any of the components of <paramref name="value"/> evaluates to positive infinity, otherwise false.</returns>
		public static bool IsPositiveInfinity(Plane value)
		{
			return
				Single.IsPositiveInfinity(value.A) ||
				Single.IsPositiveInfinity(value.B) ||
				Single.IsPositiveInfinity(value.C) ||
				Single.IsPositiveInfinity(value.D);
		}

		/// <summary>
		/// Returns a plane with the normal pointing to the opposite direction.
		/// </summary>
		/// <param name="value">A plane.</param>
		/// <returns>A plane with the normal pointing to the opposite direction.</returns>
		public static Plane Negate(Plane value)
		{
			return new Plane
			{
				A = -value.A,
				B = -value.B,
				C = -value.C,
				D = -value.D,
			};
		}

		/// <summary>
		/// Returns a plane with a normal with the magnitude of one, pointing to the same direction as the specified plane.
		/// </summary>
		/// <param name="value">A plane.</param>
		/// <returns>A plane with a normal with the magnitude of one, pointing to the same direction as the <paramref name="value"/>.</returns>
		public static Plane Normalize(Plane value)
		{
			var scale = (Single)Math.Sqrt(value.A * value.A + value.B * value.B + value.C * value.C);

			if (scale < Internals.Epsilon)
			{
				return new Plane(Vector3.UnitX, 0);
			}
			else
			{
				scale = 1 / scale;

				return new Plane
				{
					A = value.A * scale,
					B = value.B * scale,
					C = value.C * scale,
					D = value.D * scale,
				};
			}
		}

		/// <summary>
		/// Returns a plane with a normal with the magnitude of one, pointing to the same direction as the specified plane.
		/// </summary>
		/// <param name="a">The A component of the plane equation.</param>
		/// <param name="b">The B component of the plane equation.</param>
		/// <param name="c">The C component of the plane equation.</param>
		/// <param name="d">The D component of the plane equation.</param>
		/// <returns>A plane with a normal with the magnitude of one, pointing to the same direction as the specified plane.</returns>
		public static Plane Normalize(Single a, Single b, Single c, Single d)
		{
			var scale = (Single)Math.Sqrt(a * a + b * b + c * c);

			if (scale < Internals.Epsilon)
			{
				return new Plane(Vector3.UnitX, 0);
			}
			else
			{
				scale = 1 / scale;

				return new Plane
				{
					A = a * scale,
					B = b * scale,
					C = c * scale,
					D = d * scale,
				};
			}
		}

		/// <summary>
		/// Returns the plane parsed from the specified string.
		/// </summary>
		/// <param name="value">The string to parse.</param>
		/// <returns>The plane parsed from the <paramref name="value"/>.</returns>
		public static Plane Parse(string value)
		{
			return Parse(value, NumberStyles.Float | NumberStyles.AllowThousands, null);
		}

		/// <summary>
		/// Returns the plane parsed from the specified string.
		/// </summary>
		/// <param name="value">The string to parse.</param>
		/// <param name="numberStyle">The number style for each component.</param>
		/// <returns>The plane parsed from the <paramref name="value"/>.</returns>
		public static Plane Parse(string value, NumberStyles numberStyle)
		{
			return Parse(value, numberStyle, null);
		}

		/// <summary>
		/// Returns the plane parsed from the specified string.
		/// </summary>
		/// <param name="value">The string to parse.</param>
		/// <param name="formatProvider">The format provider for each component.</param>
		/// <returns>The plane parsed from the <paramref name="value"/>.</returns>
		public static Plane Parse(string value, IFormatProvider formatProvider)
		{
			return Parse(value, NumberStyles.Float | NumberStyles.AllowThousands, formatProvider);
		}

		/// <summary>
		/// Returns the plane parsed from the specified string.
		/// </summary>
		/// <param name="value">The string to parse.</param>
		/// <param name="numberStyle">The number style for each component.</param>
		/// <param name="formatProvider">The format provider for each component.</param>
		/// <returns>The plane parsed from the <paramref name="value"/>.</returns>
		public static Plane Parse(string value, NumberStyles numberStyle, IFormatProvider formatProvider)
		{
			if (value == null)
			{
				throw new ArgumentNullException("value");
			}

			var input = value.Split(Internals.Whitespace, StringSplitOptions.RemoveEmptyEntries);

			if (input.Length == ValueCount)
			{
				return new Plane
				{
					A = Single.Parse(input[0], numberStyle, formatProvider),
					B = Single.Parse(input[1], numberStyle, formatProvider),
					C = Single.Parse(input[2], numberStyle, formatProvider),
					D = Single.Parse(input[3], numberStyle, formatProvider),
				};
			}
			else
			{
				throw new FormatException("Invalid number of components in the source string.");
			}
		}

		/// <summary>
		/// Returns a string representation of the current object.
		/// </summary>
		/// <returns>A string representation of the current object.</returns>
		public override string ToString()
		{
			return ToString(null, null);
		}

		/// <summary>
		/// Returns a string representation of the current object.
		/// </summary>
		/// <param name="format">Format string.</param>
		/// <returns>A string representation of the current object.</returns>
		public string ToString(string format)
		{
			return ToString(format, null);
		}

		/// <summary>
		/// Returns a string representation of the current object.
		/// </summary>
		/// <param name="formatProvider">Format provider.</param>
		/// <returns>A string representation of the current object.</returns>
		public string ToString(IFormatProvider formatProvider)
		{
			return ToString(null, formatProvider);
		}

		/// <summary>
		/// Returns a string representation of the current object.
		/// </summary>
		/// <param name="format">Format string.</param>
		/// <param name="formatProvider">Format provider.</param>
		/// <returns>A string representation of the current object.</returns>
		public string ToString(string format, IFormatProvider formatProvider)
		{
			return Internals.Format(format, formatProvider, this.A, this.B, this.C, this.D);
		}

		/// <summary>
		/// Attempts to parse the plane from the specified string.
		/// </summary>
		/// <param name="value">The string to parse.</param>
		/// <param name="result">The output variable for the plane parsed from the <paramref name="value"/>.</param>
		/// <returns>True if the <paramref name="value"/> was parsed successfully, otherwise false.</returns>
		public static bool TryParse(string value, out Plane result)
		{
			return TryParse(value, NumberStyles.Float | NumberStyles.AllowThousands, null, out result);
		}

		/// <summary>
		/// Attempts to parse the plane from the specified string.
		/// </summary>
		/// <param name="value">The string to parse.</param>
		/// <param name="numberStyle">The number style for each component.</param>
		/// <param name="result">The output variable for the plane parsed from the <paramref name="value"/>.</param>
		/// <returns>True if the <paramref name="value"/> was parsed successfully, otherwise false.</returns>
		public static bool TryParse(string value, NumberStyles numberStyle, out Plane result)
		{
			return TryParse(value, numberStyle, null, out result);
		}

		/// <summary>
		/// Attempts to parse the plane from the specified string.
		/// </summary>
		/// <param name="value">The string to parse.</param>
		/// <param name="formatProvider">The format provider for each component.</param>
		/// <param name="result">The output variable for the plane parsed from the <paramref name="value"/>.</param>
		/// <returns>True if the <paramref name="value"/> was parsed successfully, otherwise false.</returns>
		public static bool TryParse(string value, IFormatProvider formatProvider, out Plane result)
		{
			return TryParse(value, NumberStyles.Float | NumberStyles.AllowThousands, formatProvider, out result);
		}

		/// <summary>
		/// Attempts to parse the plane from the specified string.
		/// </summary>
		/// <param name="value">The string to parse.</param>
		/// <param name="numberStyle">The number style for each component.</param>
		/// <param name="formatProvider">The format provider for each component.</param>
		/// <param name="result">The output variable for the plane parsed from the <paramref name="value"/>.</param>
		/// <returns>True if the <paramref name="value"/> was parsed successfully, otherwise false.</returns>
		public static bool TryParse(string value, NumberStyles numberStyle, IFormatProvider formatProvider, out Plane result)
		{
			if (value == null)
			{
				throw new ArgumentNullException("value");
			}

			result = new Plane();

			var input = value.Split(Internals.Whitespace, StringSplitOptions.RemoveEmptyEntries);

			return
				input.Length == ValueCount &&
				Single.TryParse(input[0], numberStyle, formatProvider, out result.A) &&
				Single.TryParse(input[1], numberStyle, formatProvider, out result.B) &&
				Single.TryParse(input[2], numberStyle, formatProvider, out result.C) &&
				Single.TryParse(input[3], numberStyle, formatProvider, out result.D);
		}

		/// <summary>
		/// Returns the current plane with the specified components replaced with the specified values.
		/// </summary>
		/// <param name="a">The new value for the <see cref="A"/> component. Use null to keep the old value.</param>
		/// <param name="b">The new value for the <see cref="B"/> component. Use null to keep the old value.</param>
		/// <param name="c">The new value for the <see cref="C"/> component. Use null to keep the old value.</param>
		/// <param name="d">The new value for the <see cref="D"/> component. Use null to keep the old value.</param>
		/// <returns>The current plane with the specified components replaced with the specified values.</returns>
		public Plane With(Single? a = null, Single? b = null, Single? c = null, Single? d = null)
		{
			return new Plane
			{
				A = a ?? this.A,
				B = b ?? this.B,
				C = c ?? this.C,
				D = d ?? this.D,
			};
		}

		/// <summary>
		/// Gets the <see cref="T:System.Xml.Schema.XmlSchema"/> that describes the XML representation of the object that is produced by the <see cref="M:System.Xml.Serialization.IXmlSerializable.WriteXml(System.Xml.XmlWriter)"/> method and consumed by the <see cref="M:System.Xml.Serialization.IXmlSerializable.ReadXml(System.Xml.XmlReader)"/> method.
		/// </summary>
		/// <returns>This method returns always null.</returns>
		System.Xml.Schema.XmlSchema IXmlSerializable.GetSchema()
		{
			return null;
		}

		/// <summary>
		/// Generates an object from its XML representation.
		/// </summary>
		/// <param name="reader">The <see cref="T:System.Xml.XmlReader"/> stream from which the object is deserialized.</param>
		void IXmlSerializable.ReadXml(XmlReader reader)
		{
			reader.MoveToContent();

			if (!reader.IsEmptyElement)
			{
				var input = reader.ReadString().Split(Internals.Whitespace, StringSplitOptions.RemoveEmptyEntries);

				if (input.Length == ValueCount)
				{
					this.A = XmlConvert.ToSingle(input[0]);
					this.B = XmlConvert.ToSingle(input[1]);
					this.C = XmlConvert.ToSingle(input[2]);
					this.D = XmlConvert.ToSingle(input[3]);
				}

				reader.ReadEndElement();
			}
		}

		/// <summary>
		/// Converts an object into its XML representation.
		/// </summary>
		/// <param name="writer">The <see cref="T:System.Xml.XmlWriter"/> stream to which the object is serialized.</param>
		void IXmlSerializable.WriteXml(XmlWriter writer)
		{
			writer.WriteString(XmlConvert.ToString(this.A));
			writer.WriteString(" ");
			writer.WriteString(XmlConvert.ToString(this.B));
			writer.WriteString(" ");
			writer.WriteString(XmlConvert.ToString(this.C));
			writer.WriteString(" ");
			writer.WriteString(XmlConvert.ToString(this.D));
		}
		#endregion

		#region Operators
		/// <summary>
		/// Determines whether the planes are equal.
		/// </summary>
		/// <param name="value1">A plane.</param>
		/// <param name="value2">A plane.</param>
		/// <returns>True if the planes are equal, otherwise false.</returns>
		public static bool operator ==(Plane value1, Plane value2)
		{
			return
				value1.A == value2.A &&
				value1.B == value2.B &&
				value1.C == value2.C &&
				value1.D == value2.D;
		}

		/// <summary>
		/// Determines whether the planes are not equal.
		/// </summary>
		/// <param name="value1">A plane.</param>
		/// <param name="value2">A plane.</param>
		/// <returns>True if the planes are not equal, otherwise false.</returns>
		public static bool operator !=(Plane value1, Plane value2)
		{
			return
				value1.A != value2.A ||
				value1.B != value2.B ||
				value1.C != value2.C ||
				value1.D != value2.D;
		}
		#endregion

		#region Indexer
		/// <summary>
		/// Gets or sets the component at the specified index.
		/// </summary>
		/// <param name="index">Component index.</param>
		/// <returns>Value of the component at the <paramref name="index"/>.</returns>
		/// <exception cref="IndexOutOfRangeException">Thrown if the <paramref name="index"/> is less than zero or greater than or equal to the number of components in the vector.</exception>
		public Single this[int index]
		{
			get
			{
				switch (index)
				{
					case 0: return this.A;
					case 1: return this.B;
					case 2: return this.C;
					case 3: return this.D;

					default:
						throw new IndexOutOfRangeException();
				}
			}
			set
			{
				switch (index)
				{
					case 0: this.A = value; break;
					case 1: this.B = value; break;
					case 2: this.C = value; break;
					case 3: this.D = value; break;

					default:
						throw new IndexOutOfRangeException();
				}
			}
		}
		#endregion

		#region Properties
		/// <summary>
		/// Gets a plane with all the components set to a value that is not a number.
		/// </summary>
		/// <value>The plane with all the components set to a value that is not a number.</value>
		public static Plane NaN
		{
			get { return nan; }
		}

		/// <summary>
		/// Gets a plane with all the components set to negative infinity.
		/// </summary>
		/// <value>The plane with all the components set to negative infinity.</value>
		public static Plane NegativeInfinity
		{
			get { return negativeInfinity; }
		}

		/// <summary>
		/// Gets or sets the plane normal vector.
		/// </summary>
		/// <value>The plane normal vector.</value>
		public Vector3 Normal
		{
			get
			{
				return new Vector3
				{
					X = this.A,
					Y = this.B,
					Z = this.C,
				};
			}
			set
			{
				this.A = value.X;
				this.B = value.Y;
				this.C = value.Z;
			}
		}

		/// <summary>
		/// Gets a plane with all the components set to positive infinity.
		/// </summary>
		/// <value>The plane with all the components set to positive infinity.</value>
		public static Plane PositiveInfinity
		{
			get { return positiveInfinity; }
		}

		/// <summary>
		/// Gets a plane with all the components set to zero.
		/// </summary>
		/// <value>The plane with all the components set to zero.</value>
		public static Plane Zero
		{
			get { return zero; }
		}
		#endregion

		#region Fields
		/// <summary>
		/// The A component of the plane equation.
		/// </summary>
		public Single A;

		/// <summary>
		/// The B component of the plane equation.
		/// </summary>
		public Single B;

		/// <summary>
		/// The C component of the plane equation.
		/// </summary>
		public Single C;

		/// <summary>
		/// The D component of the plane equation.
		/// </summary>
		public Single D;

		/// <summary>
		/// The number of values in the plane equation.
		/// </summary>
		public const int ValueCount = 4;

		/// <summary>
		/// A plane with all the components set to a value that is not a number.
		/// </summary>
		private static readonly Plane nan = new Plane(Single.NaN);

		/// <summary>
		/// A plane with all the components set to negative infinity.
		/// </summary>
		private static readonly Plane negativeInfinity = new Plane(Single.NegativeInfinity);

		/// <summary>
		/// A plane with all the components set to positive infinity.
		/// </summary>
		private static readonly Plane positiveInfinity = new Plane(Single.PositiveInfinity);

		/// <summary>
		/// A plane with all the components set to zero.
		/// </summary>
		private static readonly Plane zero = new Plane(0);
		#endregion
	}
}
